/* Platform */
#include "vl53l1_platform.h"

#define PLATFORM_MAXIMUM_MULTI_WRITE_SIZE 256

VL53L1_Error VL53L1_WaitUs(VL53L1_Dev_t *pdev, int32_t wait_us) {
    return pdev->Ops.Delay_Us(pdev->OpsData, wait_us);
}

VL53L1_Error VL53L1_WaitMs(VL53L1_Dev_t *pdev, int32_t wait_ms) {
    return pdev->Ops.Delay_Ms(pdev->OpsData, wait_ms);
}

VL53L1_Error VL53L1_RdByte(VL53L1_Dev_t *pdev, uint16_t index, uint8_t *pdata) {
    uint8_t tx_buf[2];

    tx_buf[0] = (index >> 8U) & 0xFFU;
    tx_buf[1] = (index & 0xFFU);

    VL53L1_Dev_Xfer_Desc_t xfer = {
        .TxData = tx_buf,
        .TxSize = 2,
        .RxData = pdata,
        .RxSize = 1,
    };

    return pdev->Ops.I2C_Xfer(pdev->OpsData, &xfer);
}

VL53L1_Error VL53L1_RdWord(VL53L1_Dev_t *pdev, uint16_t index, uint16_t *pdata) {
    uint8_t tx_buf[2];
    uint8_t rx_buf[2];

    tx_buf[0] = (index >> 8U) & 0xFFU;
    tx_buf[1] = (index & 0xFFU);

    VL53L1_Dev_Xfer_Desc_t xfer = {
        .TxData = tx_buf,
        .TxSize = 2,
        .RxData = rx_buf,
        .RxSize = 2,
    };

    VL53L1_Error ret = pdev->Ops.I2C_Xfer(pdev->OpsData, &xfer);
    if (ret != VL53L1_ERROR_NONE) return ret;

    *pdata = (rx_buf[0] << 8U) | rx_buf[1];

    return ret;
}

VL53L1_Error VL53L1_ReadMulti(VL53L1_Dev_t *pdev, uint16_t index, uint8_t *pdata, uint32_t count) {
    uint8_t tx_buf[2];

    tx_buf[0] = (index >> 8U) & 0xFFU;
    tx_buf[1] = (index & 0xFFU);

    VL53L1_Dev_Xfer_Desc_t xfer = {
        .TxData = tx_buf,
        .TxSize = 2,
        .RxData = pdata,
        .RxSize = count,
    };

    return pdev->Ops.I2C_Xfer(pdev->OpsData, &xfer);
}

VL53L1_Error VL53L1_WrByte(VL53L1_Dev_t *pdev, uint16_t index, uint8_t data) {
    uint8_t tx_buf[3];

    tx_buf[0] = (index >> 8U) & 0xFFU;
    tx_buf[1] = (index & 0xFFU);
    tx_buf[2] = data;

    VL53L1_Dev_Xfer_Desc_t xfer = {
        .TxData = tx_buf,
        .TxSize = 3,
        .RxData = NULL,
        .RxSize = 0,
    };

    return pdev->Ops.I2C_Xfer(pdev->OpsData, &xfer);
}

VL53L1_Error VL53L1_WriteMulti(VL53L1_Dev_t *pdev, uint16_t index, uint8_t *pdata, uint32_t count) {
    if (count > PLATFORM_MAXIMUM_MULTI_WRITE_SIZE - 2) {
        for (;;) {
            /* -- */
            pdev->Ops.Delay_Ms(pdev->OpsData, 1000);
        }
    }

    uint8_t tx_buf[PLATFORM_MAXIMUM_MULTI_WRITE_SIZE];

    tx_buf[0] = (index >> 8U) & 0xFFU;
    tx_buf[1] = (index & 0xFFU);

    memcpy(&tx_buf[2], pdata, count);

    VL53L1_Dev_Xfer_Desc_t xfer = {
        .TxData = tx_buf,
        .TxSize = count + 2,
        .RxData = NULL,
        .RxSize = 0,
    };

    return pdev->Ops.I2C_Xfer(pdev->OpsData, &xfer);
}

VL53L1_Error VL53L1_WaitValueMaskEx(VL53L1_Dev_t *pdev, uint32_t timeout_ms, uint16_t index, uint8_t value,
                                    uint8_t mask, uint32_t poll_delay_ms) {
    uint8_t tx_buf[2];
    uint8_t rx_buf[1];

    tx_buf[0] = (index >> 8U) & 0xFFU;
    tx_buf[1] = (index & 0xFFU);

    VL53L1_Dev_Xfer_Desc_t xfer = {
        .TxData = tx_buf,
        .TxSize = 2,
        .RxData = rx_buf,
        .RxSize = 1,
    };

    uint8_t ts_count = 0U;

    VL53L1_Error ret;

    do {
        ret = pdev->Ops.I2C_Xfer(pdev->OpsData, &xfer);
        if (ret != VL53L1_ERROR_NONE) return ret;

        ts_count += poll_delay_ms;

        pdev->Ops.Delay_Ms(pdev->OpsData, poll_delay_ms);

        if (ts_count > timeout_ms) {
            return VL53L1_ERROR_TIME_OUT;
        }
    } while ((rx_buf[0] & mask) != value);

    return VL53L1_ERROR_NONE;
}
